-----------------------------------
--         Oneko Cat 1.0         --
--      by Unmatched Bracket     --
-- Assets from other Oneko ports --
--     Reusable, give credit     --
-----------------------------------

local STATE = {
    IDLE = "IDLE",
    MOVING = "MOVING",
    MLEM = "MLEM",
    ITCH = "ITCH",
    CLAW = "CLAW",
    YAWN = "YAWN",
    SLEP = "SLEP"
}
local ONEKOSIZE = 32
local ONEKOBEAT = 6
local DIRS = {
    LEFT = 1,
    RIGHT = 2,
    UP = 4,
    DOWN = 8,
    [1] = "LEFT",
    [2] = "RIGHT",
    [4] = "UP",
    [5] = "UPLEFT",
    [6] = "UPRIGHT",
    [8] = "DOWN",
    [9] = "DOWNLEFT",
    [10] = "DOWNRIGHT"
}

local function newoneko()
    return {
        x = 100,
        y = 100,
        state = STATE.IDLE,
        statedata = nil,
        ticker = 35,
        eepy = 0
    }
end
local onekii = {}
local onekospawncool = 0

local cv_onekocount = CV_RegisterVar({
    name = "oneko",
    defaultvalue = "1",
    PossibleValue = {MIN = 0, MAX = 100}
})

--[[@param v videolib]]
local function specialMoveTarget(oneko, v)
    if v.RandomChance(FU/20) then
        local key = v.RandomRange(1,4)
        local snap = ({
            {"x", 0, "LEFT"},
            {"x", v.width()-ONEKOSIZE, "RIGHT"},
            {"y", 0, "UP"},
            {"y", v.height()-ONEKOSIZE, "DOWN"},
        })[key]
        oneko.statedata[snap[1]] = snap[2]
        oneko.statedata.snapdir = snap[3]
    elseif v.RandomChance(FU/10) and not camera.chase and not splitscreen then
        local crossx = (v.width()-ONEKOSIZE)/2
        local crossy = (v.height()/2)-ONEKOSIZE+4
        -- don't sit on the crosshair if anyone else is
        for _, o in pairs(onekii) do
            if (o.x == crossx and o.y == crossy)
            or (o.state == STATE.MOVING and o.statedata
                and o.statedata.x == crossx and o.statedata.y == crossy)
            then
                return
            end
        end
        -- sit on the crosshair
        oneko.statedata.x = crossx
        oneko.statedata.y = crossy
    end
end

--[[@param v videolib]]
local function tickOneko(oneko, v)
    if not oneko then return end

    if oneko.state == STATE.MOVING then
        local dir = 0

        -- straight line bonus for consistient speed
        local straight = 0
        if oneko.x == oneko.statedata.x
        or oneko.y == oneko.statedata.y
        then
            straight = 4*ONEKOBEAT/10
        end

        for _=1,(ONEKOBEAT+straight) do
            dir = 0
            if oneko.x > oneko.statedata.x then
                dir = $ | DIRS.LEFT
                oneko.x = $ - 1
            elseif oneko.x < oneko.statedata.x then
                dir = $ | DIRS.RIGHT
                oneko.x = $ + 1
            end
            if oneko.y > oneko.statedata.y then
                dir = $ | DIRS.UP
                oneko.y = $ - 1
            elseif oneko.y < oneko.statedata.y then
                dir = $ | DIRS.DOWN
                oneko.y = $ + 1
            end
            if dir ~= 0 then
                oneko.statedata.dir = dir
            end
        end
        if dir == 0 then
            oneko.ticker = 0
        end
    end

    oneko.ticker = $ - 1
    local newstate = nil
    if oneko.ticker <= 0 then
        if oneko.state == STATE.IDLE then
            local opt = {
                STATE.MOVING, STATE.MOVING, STATE.MOVING, STATE.MOVING,
                STATE.MLEM, STATE.MLEM, STATE.MLEM,
                STATE.ITCH, STATE.ITCH, STATE.ITCH,
                STATE.YAWN, STATE.YAWN
            }
            newstate = opt[v.RandomRange(1, #opt)]
        elseif
            oneko.state == STATE.MOVING and oneko.statedata.snapdir
            and oneko.x == oneko.statedata.x and oneko.y == oneko.statedata.y
        then
            newstate = STATE.CLAW
        elseif
            oneko.state == STATE.YAWN and oneko.eepy > 2
            and v.RandomChance(FU/min(1, 5-oneko.eepy))
        then
            newstate = STATE.SLEP
        else
            newstate = STATE.IDLE
        end
    end

    if newstate then
        if newstate == STATE.IDLE then
            oneko.ticker = v.RandomRange(70/ONEKOBEAT, 175/ONEKOBEAT)
            oneko.statedata = nil
        elseif newstate == STATE.MOVING then
            oneko.statedata = {
                x = v.RandomRange(0, v.width()-ONEKOSIZE),
                y = v.RandomRange(0, v.height()-ONEKOSIZE),
                dir = -1
            }
            specialMoveTarget(oneko, v)
            oneko.ticker = max(
                abs(oneko.x - oneko.statedata.x),
                abs(oneko.y - oneko.statedata.y)
            )+5
        elseif newstate == STATE.MLEM then
            oneko.ticker = 4*v.RandomRange(3, 5)
            oneko.statedata = nil
        elseif newstate == STATE.ITCH then
            oneko.ticker = 2*v.RandomRange(5, 13)
            oneko.statedata = nil
        elseif newstate == STATE.CLAW then
            oneko.ticker = 2*v.RandomRange(5, 13)
            oneko.statedata = {
                dir = oneko.statedata.snapdir
            }
        elseif newstate == STATE.YAWN then
            oneko.ticker = 4*v.RandomRange(2, 3)
            oneko.eepy = $ + 1
        elseif newstate == STATE.SLEP then
            oneko.ticker = 8*v.RandomRange(7, 16)
            oneko.eepy = 0
        end
        oneko.state = newstate
    end
end

local lastw = nil
local lasth = nil
local function remap(x, prev, new)
    return x*(new-ONEKOSIZE)/(prev-ONEKOSIZE)
end
local function handleResize(oneko, neww, newh)
    if lastw == nil or lasth == nil then return end
    oneko.x = remap($, lastw, neww)
    oneko.y = remap($, lasth, newh)
    if oneko.state == STATE.MOVING and oneko.statedata then
        oneko.statedata.x = remap($, lastw, neww)
        oneko.statedata.y = remap($, lasth, newh)
        oneko.ticker = max(
            abs(oneko.x - oneko.statedata.x),
            abs(oneko.y - oneko.statedata.y)
        )+5
    end
end

local function drawoneko(oneko, v)
    local sprite = "YAWN1" -- idle
    local beat = oneko.ticker+8
    if oneko.state == STATE.MOVING then
        local dir = DIRS[oneko.statedata.dir]
        if dir then
            sprite = dir .. (beat%2+1)
        end
    elseif oneko.state == STATE.CLAW then
        local dir = oneko.statedata.dir
        if DIRS[dir] then
            sprite = dir .. "CLAW" .. (beat%2+1)
        end
    elseif oneko.state == STATE.MLEM then
        sprite = "WASH" .. ((beat/2)%2+1)
    elseif oneko.state == STATE.ITCH then
        sprite = "SCRATCH" .. (beat%2+1)
    elseif oneko.state == STATE.YAWN then
        sprite = "YAWN" .. ((beat/2)%2+1)
    elseif oneko.state == STATE.SLEP then
        sprite = "SLEEP" .. ((beat/4)%2+1)
    end
    v.draw(
        oneko.x,
        oneko.y,
        v.cachePatch("ONEKO_" .. sprite),
        V_NOSCALEPATCH|V_SNAPTOLEFT|V_SNAPTOTOP
    )
end

local tick = 0
--[[@param v videolib]]
local function drawer (v)
    local w, h = v.width(), v.height()
    tick = $ + 1
    if w ~= lastw or h ~= lasth then
        -- pcall(function ()
        --     CV_Set(cv_onekocount, w * h / ONEKOSIZE / ONEKOSIZE * 2)
        -- end)
        for _,oneko in pairs(onekii) do
            pcall(handleResize, oneko, w, h)
        end
        lastw = w
        lasth = h
    end
    if cv_onekocount.value > #onekii and onekospawncool <= 0 then
        local newcats = 1 + (cv_onekocount.value - #onekii) / 20
        for _=1,newcats do
            onekospawncool = ONEKOBEAT
            local new = newoneko()
            new.x = v.RandomRange(0, v.width() - ONEKOSIZE)
            new.y = -ONEKOSIZE
            new.state = STATE.MOVING
            new.statedata = {
                x = new.x,
                y = v.RandomRange(0, v.height()-ONEKOSIZE),
                dir = 0
            }
            new.ticker = new.statedata.y + 16 + ONEKOSIZE
            table.insert(onekii, new)
        end
    elseif cv_onekocount.value < #onekii then
        local i = cv_onekocount.value+1
        while i <= #onekii do
            local oneko = onekii[i]
            if oneko.y < -ONEKOSIZE then
                table.remove(onekii, i)
                continue
            elseif tick >= ONEKOBEAT and v.RandomRange(i, #onekii) >= (#onekii-10) then
                oneko.state = STATE.MOVING
                if not (
                    oneko.statedata and oneko.statedata.y
                    and oneko.statedata.y < -ONEKOSIZE
                ) then
                    oneko.statedata = {dir = 0}
                    oneko.statedata.x = oneko.x
                    oneko.statedata.y = -ONEKOSIZE*2
                end
                if oneko.ticker < 1000 then
                    oneko.ticker = $ + ONEKOBEAT*2
                end
            end
            i = $ + 1
        end
    end
    if onekospawncool > 0 then
        onekospawncool = $ - 1
    end
    for _,oneko in pairs(onekii) do
        if tick >= ONEKOBEAT then
            pcall(tickOneko, oneko, v)
        end
        pcall(drawoneko, oneko, v)
    end
    if tick >= ONEKOBEAT then
        tick = 0
    end
end

hud.add(drawer, "game")
hud.add(drawer, "scores")
hud.add(drawer, "intermission")
hud.add(drawer, "title")
pcall(hud.add, drawer, "continue")
pcall(hud.add, drawer, "playersetup")